import unittest
import numpy as np

from src.sequencer import Sequencer
from src.director import Director


class TestSequencer(unittest.TestCase):
    """
    Test sequencer methods
    """

    def setup_dependencies(self):
        self.director = Director()
        self.sequencer = Sequencer(director=self.director)
        # TODO add these outputs to a fixtures file
        # single value change - 1 channel, 1 item in sequence
        self.one_channel_output = [
            {
                "sequence": [
                    {
                        "back": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "front": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "bottom": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "top": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "right": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "left": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "middle": (0.04000000000000001, 0.6400000000000001, 0.76),
                    }
                ],
                "weight": 0.1,
                "origin": "pattern_sequencer",
            }
        ]

        # two stacked sequences - 1 length seq from pattern sequencer, bigger seq from gesture
        self.two_channel_output = [
            {
                "sequence": [
                    {
                        "back": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "front": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "bottom": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "top": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "right": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "left": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "middle": (0.04000000000000001, 0.6400000000000001, 0.76),
                    }
                ],
                "weight": 0.1,
                "origin": "pattern_sequencer",
            },
            {
                "sequence": [
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.0, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.0, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.00784313725490196, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.00392156862745098, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.0196078431372549, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.00784313725490196, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.03137254901960784, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.01568627450980392, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.043137254901960784, 0.0),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.0196078431372549, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.054901960784313725, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.027450980392156862, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.06274509803921569, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.03137254901960784, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.07058823529411765, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.0392156862745098, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.08235294117647059, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.047058823529411764, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.10196078431372549, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.054901960784313725, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.13725490196078433, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.07450980392156863, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.19607843137254902, 0.00784313725490196),
                        "left": (0.0, 0.01568627450980392, 0.0),
                        "right": (0.0, 0.09803921568627451, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.00392156862745098, 0.0),
                        "bottom": (0.0, 0.27058823529411763, 0.011764705882352941),
                        "left": (0.0, 0.03137254901960784, 0.0),
                        "right": (0.0, 0.13333333333333333, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.011764705882352941, 0.0),
                        "bottom": (0.0, 0.3764705882352941, 0.01568627450980392),
                        "left": (0.0, 0.058823529411764705, 0.00392156862745098),
                        "right": (0.0, 0.17647058823529413, 0.00784313725490196),
                    },
                    {
                        "top": (0.0, 0.023529411764705882, 0.0),
                        "bottom": (
                            0.00392156862745098,
                            0.5176470588235295,
                            0.0196078431372549,
                        ),
                        "left": (0.0, 0.09411764705882353, 0.00392156862745098),
                        "right": (0.0, 0.23529411764705882, 0.00784313725490196),
                    },
                    {
                        "top": (0.0, 0.043137254901960784, 0.0),
                        "bottom": (
                            0.00392156862745098,
                            0.6588235294117647,
                            0.023529411764705882,
                        ),
                        "left": (0.0, 0.1411764705882353, 0.00392156862745098),
                        "right": (0.0, 0.30196078431372547, 0.011764705882352941),
                    },
                    {
                        "top": (0.0, 0.06666666666666667, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.7843137254901961,
                            0.027450980392156862,
                        ),
                        "left": (0.0, 0.19215686274509805, 0.00784313725490196),
                        "right": (0.0, 0.3607843137254902, 0.011764705882352941),
                    },
                    {
                        "top": (0.0, 0.10196078431372549, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "left": (0.0, 0.24705882352941178, 0.00784313725490196),
                        "right": (0.0, 0.4196078431372549, 0.01568627450980392),
                    },
                    {
                        "top": (0.0, 0.1411764705882353, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.9411764705882353,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.2980392156862745, 0.011764705882352941),
                        "right": (0.0, 0.47058823529411764, 0.01568627450980392),
                    },
                    {
                        "top": (0.0, 0.19215686274509805, 0.00784313725490196),
                        "bottom": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.34901960784313724, 0.011764705882352941),
                        "right": (
                            0.00392156862745098,
                            0.5098039215686274,
                            0.0196078431372549,
                        ),
                    },
                    {
                        "top": (0.0, 0.23921568627450981, 0.00784313725490196),
                        "bottom": (
                            0.00392156862745098,
                            0.9450980392156862,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.39215686274509803, 0.01568627450980392),
                        "right": (
                            0.00392156862745098,
                            0.5372549019607843,
                            0.0196078431372549,
                        ),
                    },
                    {
                        "top": (0.0, 0.3686274509803922, 0.011764705882352941),
                        "bottom": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.49411764705882355, 0.0196078431372549),
                        "right": (
                            0.00392156862745098,
                            0.6313725490196078,
                            0.023529411764705882,
                        ),
                    },
                    {
                        "top": (0.0, 0.4980392156862745, 0.0196078431372549),
                        "bottom": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03529411764705882,
                        ),
                        "left": (0.00392156862745098, 0.6, 0.023529411764705882),
                        "right": (
                            0.00392156862745098,
                            0.7215686274509804,
                            0.027450980392156862,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.6274509803921569,
                            0.023529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.7019607843137254,
                            0.027450980392156862,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.807843137254902,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.7450980392156863,
                            0.027450980392156862,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8431372549019608,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.803921568627451,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.788235294117647,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9176470588235294,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7607843137254902,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9568627450980393,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7333333333333333,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9176470588235294,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9490196078431372,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7058823529411765,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9058823529411765,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9372549019607843,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6784313725490196,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9254901960784314,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8784313725490196,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6509803921568628,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9137254901960784,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8745098039215686,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6235294117647059,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9019607843137255,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.592156862745098,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5647058823529412,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8627450980392157,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5372549019607843,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8352941176470589,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5098039215686274,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8235294117647058,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4823529411764706, 0.0196078431372549),
                        "left": (
                            0.00392156862745098,
                            0.8196078431372549,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8627450980392157,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4549019607843137, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.42745098039215684, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8196078431372549,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8235294117647058,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9058823529411765,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.37254901960784315, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8352941176470589,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8431372549019608,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9294117647058824,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.34509803921568627, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3176470588235294, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8549019607843137,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3254901960784314, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3686274509803922, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.43529411764705883, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9333333333333333,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5176470588235295,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9294117647058824,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6235294117647059,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9725490196078431,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9568627450980393,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7019607843137254,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9450980392156862,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7647058823529411,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9333333333333333,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9921568627450981,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8549019607843137,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9921568627450981,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8784313725490196,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.7098039215686275,
                            0.027450980392156862,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7215686274509804,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.7843137254901961,
                            0.027450980392156862,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.7803921568627451,
                            0.027450980392156862,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.5294117647058824,
                            0.0196078431372549,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5411764705882353,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.5843137254901961,
                            0.023529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.5843137254901961,
                            0.023529411764705882,
                        ),
                    },
                ],
                "weight": 0.99,
                "origin": "gesture_pipeline",
            },
        ]

        self.three_channel_output = [
            {
                "sequence": [
                    {
                        "back": (0.0, 0.0, 0.0),
                        "front": (1.0, 1.0, 1.0),
                        "bottom": (1.0, 1.0, 1.0),
                        "top": (0.0, 0.0, 0.0),
                        "right": (0.3, 0.3, 0.3),
                        "left": (0.7, 0.7, 0.7),
                        "middle": (0.5, 0.5, 0.5),
                    }
                ],
                "weight": 0.5,
                "origin": "fuzzy_tracker",
            },
            {
                "sequence": [
                    {
                        "back": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "front": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "bottom": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "top": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "right": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "left": (0.04000000000000001, 0.6400000000000001, 0.76),
                        "middle": (0.04000000000000001, 0.6400000000000001, 0.76),
                    }
                ],
                "weight": 0.1,
                "origin": "pattern_sequencer",
            },
            {
                "sequence": [
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.0, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.0, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.00784313725490196, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.00392156862745098, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.0196078431372549, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.00784313725490196, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.03137254901960784, 0.0),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.01568627450980392, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.043137254901960784, 0.0),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.0196078431372549, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.054901960784313725, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.027450980392156862, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.06274509803921569, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.03137254901960784, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.07058823529411765, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.0392156862745098, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.08235294117647059, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.047058823529411764, 0.0),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.10196078431372549, 0.00392156862745098),
                        "left": (0.0, 0.0, 0.0),
                        "right": (0.0, 0.054901960784313725, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.13725490196078433, 0.00392156862745098),
                        "left": (0.0, 0.00392156862745098, 0.0),
                        "right": (0.0, 0.07450980392156863, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.0, 0.0),
                        "bottom": (0.0, 0.19607843137254902, 0.00784313725490196),
                        "left": (0.0, 0.01568627450980392, 0.0),
                        "right": (0.0, 0.09803921568627451, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.00392156862745098, 0.0),
                        "bottom": (0.0, 0.27058823529411763, 0.011764705882352941),
                        "left": (0.0, 0.03137254901960784, 0.0),
                        "right": (0.0, 0.13333333333333333, 0.00392156862745098),
                    },
                    {
                        "top": (0.0, 0.011764705882352941, 0.0),
                        "bottom": (0.0, 0.3764705882352941, 0.01568627450980392),
                        "left": (0.0, 0.058823529411764705, 0.00392156862745098),
                        "right": (0.0, 0.17647058823529413, 0.00784313725490196),
                    },
                    {
                        "top": (0.0, 0.023529411764705882, 0.0),
                        "bottom": (
                            0.00392156862745098,
                            0.5176470588235295,
                            0.0196078431372549,
                        ),
                        "left": (0.0, 0.09411764705882353, 0.00392156862745098),
                        "right": (0.0, 0.23529411764705882, 0.00784313725490196),
                    },
                    {
                        "top": (0.0, 0.043137254901960784, 0.0),
                        "bottom": (
                            0.00392156862745098,
                            0.6588235294117647,
                            0.023529411764705882,
                        ),
                        "left": (0.0, 0.1411764705882353, 0.00392156862745098),
                        "right": (0.0, 0.30196078431372547, 0.011764705882352941),
                    },
                    {
                        "top": (0.0, 0.06666666666666667, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.7843137254901961,
                            0.027450980392156862,
                        ),
                        "left": (0.0, 0.19215686274509805, 0.00784313725490196),
                        "right": (0.0, 0.3607843137254902, 0.011764705882352941),
                    },
                    {
                        "top": (0.0, 0.10196078431372549, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "left": (0.0, 0.24705882352941178, 0.00784313725490196),
                        "right": (0.0, 0.4196078431372549, 0.01568627450980392),
                    },
                    {
                        "top": (0.0, 0.1411764705882353, 0.00392156862745098),
                        "bottom": (
                            0.00392156862745098,
                            0.9411764705882353,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.2980392156862745, 0.011764705882352941),
                        "right": (0.0, 0.47058823529411764, 0.01568627450980392),
                    },
                    {
                        "top": (0.0, 0.19215686274509805, 0.00784313725490196),
                        "bottom": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.34901960784313724, 0.011764705882352941),
                        "right": (
                            0.00392156862745098,
                            0.5098039215686274,
                            0.0196078431372549,
                        ),
                    },
                    {
                        "top": (0.0, 0.23921568627450981, 0.00784313725490196),
                        "bottom": (
                            0.00392156862745098,
                            0.9450980392156862,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.39215686274509803, 0.01568627450980392),
                        "right": (
                            0.00392156862745098,
                            0.5372549019607843,
                            0.0196078431372549,
                        ),
                    },
                    {
                        "top": (0.0, 0.3686274509803922, 0.011764705882352941),
                        "bottom": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "left": (0.0, 0.49411764705882355, 0.0196078431372549),
                        "right": (
                            0.00392156862745098,
                            0.6313725490196078,
                            0.023529411764705882,
                        ),
                    },
                    {
                        "top": (0.0, 0.4980392156862745, 0.0196078431372549),
                        "bottom": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03529411764705882,
                        ),
                        "left": (0.00392156862745098, 0.6, 0.023529411764705882),
                        "right": (
                            0.00392156862745098,
                            0.7215686274509804,
                            0.027450980392156862,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.6274509803921569,
                            0.023529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.7019607843137254,
                            0.027450980392156862,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.807843137254902,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.7450980392156863,
                            0.027450980392156862,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8431372549019608,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.803921568627451,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.788235294117647,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9176470588235294,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9607843137254902,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7607843137254902,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9568627450980393,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7333333333333333,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9176470588235294,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9490196078431372,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7058823529411765,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9058823529411765,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9372549019607843,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6784313725490196,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9254901960784314,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8784313725490196,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6509803921568628,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9137254901960784,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8745098039215686,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6235294117647059,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9019607843137255,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.592156862745098,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5647058823529412,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8823529411764706,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8627450980392157,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5372549019607843,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8352941176470589,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5098039215686274,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.8235294117647058,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4823529411764706, 0.0196078431372549),
                        "left": (
                            0.00392156862745098,
                            0.8196078431372549,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8627450980392157,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4549019607843137, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.42745098039215684, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8196078431372549,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                        "bottom": (0.0, 0.4, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8235294117647058,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8392156862745098,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9058823529411765,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.37254901960784315, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8352941176470589,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8431372549019608,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9294117647058824,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.34509803921568627, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8470588235294118,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8588235294117647,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3176470588235294, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8549019607843137,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8666666666666667,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3254901960784314, 0.011764705882352941),
                        "left": (
                            0.00392156862745098,
                            0.8705882352941177,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.8862745098039215,
                            0.03137254901960784,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.3686274509803922, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.8901960784313725,
                            0.03137254901960784,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "bottom": (0.0, 0.43529411764705883, 0.01568627450980392),
                        "left": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9333333333333333,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5176470588235295,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9294117647058824,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.6235294117647059,
                            0.023529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9529411764705882,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9725490196078431,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9568627450980393,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7019607843137254,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9686274509803922,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9450980392156862,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7647058823529411,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9333333333333333,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8156862745098039,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9921568627450981,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9215686274509803,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8549019607843137,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9921568627450981,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.9098039215686274,
                            0.03529411764705882,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8784313725490196,
                            0.03137254901960784,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.9882352941176471,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03137254901960784,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.8941176470588236,
                            0.03529411764705882,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.984313725490196,
                            0.03529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.9803921568627451,
                            0.03529411764705882,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.7098039215686275,
                            0.027450980392156862,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.7215686274509804,
                            0.027450980392156862,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.7843137254901961,
                            0.027450980392156862,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.7803921568627451,
                            0.027450980392156862,
                        ),
                    },
                    {
                        "top": (
                            0.00392156862745098,
                            0.5294117647058824,
                            0.0196078431372549,
                        ),
                        "bottom": (
                            0.00392156862745098,
                            0.5411764705882353,
                            0.0196078431372549,
                        ),
                        "left": (
                            0.00392156862745098,
                            0.5843137254901961,
                            0.023529411764705882,
                        ),
                        "right": (
                            0.00392156862745098,
                            0.5843137254901961,
                            0.023529411764705882,
                        ),
                    },
                ],
                "weight": 0.99,
                "origin": "gesture_pipeline",
            },
        ]
        self.spatial_categories = [
            "left",
            "right",
            "top",
            "bottom",
            "back",
            "front",
            "middle",
        ]

    def teardown_dependencies(self):
        self.director = None
        self.sequencer = None
        self.one_channel_output = []
        self.two_channel_output = []
        self.three_channel_output = []

    def test_empty_outputs_to_sequencer(self):
        self.setup_dependencies()
        result = self.sequencer.add_output_sequences_to_queue(outputs=[])
        self.assertEqual(result, None)
        self.teardown_dependencies()

    def test_sequencer_weights(self):
        """
        All outputs should have a weight attached.
        Relative to one another the weights should all sum to 1.
        One output weighted .1 should result in summation to 1
        Two outputs weighted .1 and .2 respectively should still
        be normalized and sum to 1 where the latter weight is still
        twice as large as the former, etc
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            weights = self.sequencer.current_sequence_weights
            results.append(np.allclose(np.sum(weights), 1, rtol=1e-05, atol=1e-08))
            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()
        assert all(results)

    def test_sequencer_queue_lengths(self):
        """
        The queue should end up being as long as the longest output sequence passed
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            max_sequence_len = 0
            for seq in self.sequencer.current_sequences:
                if len(seq) > max_sequence_len:
                    max_sequence_len = len(seq)
            queue = self.sequencer.queue
            self.assertEqual(len(queue), max_sequence_len)
            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()
        assert all(results)

    def test_unmerged_queue_weights(self):
        """
        Where there are multiple outputs, the length of each frame
        should correspond to the number of channels active in that
        queue position. The weights for each frame should sum to 1.
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            # the weights for every frame should sum to 1
            weight_results = []
            for frame in self.sequencer.unmerged_queue:
                weights = []
                for seq in frame:
                    weights.append(seq["weight"])
                weight_sum = np.sum(weights)
                weight_results.append(np.allclose(weight_sum, 1))
            results.append(all(weight_results))

            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()
        assert all(results)

    def test_merged_queue_values(self):
        """
        Each sequence frame value for each color should be mixed together according
        to the corresponding sequence weight. If there are existing queue values,
        they should also be taken into account
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        ch = 1
        pos = "top"
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            values = []
            for i, frame in enumerate(self.sequencer.unmerged_queue):
                top_value = 0.0
                for seq in frame:
                    # compare 1 channel value
                    top_value += np.round(seq["weight"] * seq[pos][ch], 3)
                values.append(top_value)
            queue_values = []
            for i, v in enumerate(values):
                self.sequencer.queue[i]["top"]
                queue_values.append(self.sequencer.queue[i][pos][ch])
            results.append(values == queue_values)
            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()
        assert all(results)

    def test_valid_queue_values(self):
        """
        All queue values should be between 0-1
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            for frame in self.sequencer.queue:
                for pos in self.spatial_categories:
                    if pos in frame:
                        for ch in range(3):
                            self.assertGreaterEqual(frame[pos][ch], 0)
                            self.assertLessEqual(frame[pos][ch], 1)
            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()

    def test_adding_multiple_sequences(self):
        """
        If there are preexisting sequence frames we are layering new
        sequences on top of, make sure they are valid values
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)

            # let's move deeper into the queue and add in additional sequences
            for i in range(15):
                self.sequencer.queue.popleft()
                self.sequencer.add_output_sequences_to_queue(self.three_channel_output)
            for i, _ in enumerate(self.sequencer.queue):
                for ch in range(3):
                    for pos in self.spatial_categories:
                        if pos in self.sequencer.queue[i]:
                            self.assertGreaterEqual(self.sequencer.queue[i][pos][ch], 0)
                            self.assertLessEqual(self.sequencer.queue[i][pos][ch], 1)
            self.teardown_dependencies()
            self.setup_dependencies()

    def test_director_pass_thru(self):
        """
        The sequence passes the current queue to the director
        They should be equal at the outset
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            self.assertEqual(self.sequencer.queue, self.director.current_queue)
            self.assertEqual(
                self.sequencer.queue_meta, self.director.current_queue_meta
            )
            self.sequencer.get_next_values()
            self.assertEqual(self.sequencer.queue, self.director.current_queue)
            self.assertEqual(
                self.sequencer.queue_meta, self.director.current_queue_meta
            )
            self.teardown_dependencies()
            self.setup_dependencies()
        self.teardown_dependencies()

    def test_queue_length(self):
        """
        The length of the queue should decrease by one when getting next values
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            while len(self.sequencer.queue):
                queue_length = len(self.sequencer.queue)
                self.sequencer.get_next_values()
                self.assertEqual(queue_length - 1, len(self.sequencer.queue))
            self.teardown_dependencies()
            self.setup_dependencies()
        self.assertEqual(True, all(results))
        self.teardown_dependencies()

    def test_next_values(self):
        """
        The next values should match the expected next values
        Next values should remove those values from the queue
        """
        self.setup_dependencies()
        outputs = [
            self.one_channel_output,
            self.two_channel_output,
            self.three_channel_output,
        ]
        results = []
        for o in outputs:
            self.sequencer.add_output_sequences_to_queue(o)
            self.assertEqual(self.sequencer.queue, self.director.current_queue)
            while len(self.sequencer.queue):
                queue_values = self.sequencer.queue[0]
                next_values = self.sequencer.get_next_values()
                if next_values:
                    for k in next_values.keys():
                        results.append(queue_values[k] == next_values[k])
            self.teardown_dependencies()
            self.setup_dependencies()
        self.assertEqual(True, all(results))
        self.teardown_dependencies()


if __name__ == "__main__":
    unittest.main()
